/*
  smtppipe - sends mail from stdin to a smtp relay
  see also: http://www.ogris.de/smtppipe/

  smtppipe --help gives you detailed help
  or look at usage() below

  this software is freely distributable
  as described by the terms of the GNU GPL;
  see http://www.gnu.org/licenses/gpl.html
*/

#include <unistd.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <netdb.h>
#include <resolv.h>
#include <fcntl.h>
#include <errno.h>
#include <pwd.h>
#include <signal.h>
#include <time.h>
#include "version.h"

#define STDIN 0
#define STDOUT 1
#define STDERR 2

#define BLKSIZE 4096       /* buffers are allocated in multiples of BLKSIZE */
#define SMTP_PORT 25              /* maybe overwritten by cmd line argument */
#define NULLSTRING { NULL, 0, 0 } /* initialization for Strings             */
#define CONNECT_TIMEOUT 60        /* maybe overwritten by cmd line argument */
#define READ_TIMEOUT 120          /* "                                      */
#define WRITE_TIMEOUT 120         /* "                                      */

#define INFOP(a,b) if (verbosity >= 1) outp(STDOUT, a, b, "INFO: ")
#define DEBUGP(a,b) if (verbosity >= 2) outp(STDOUT, a, b, "DEBUG: ");

/* for generating Message-ID and Date */
char *months[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun",
                  "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
char *days[] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};


typedef struct {
  char *s;   /* zero terminated                                  */
  size_t l;  /* length of string hold in s                       */
  size_t _l; /* length of buffer allocated; multiples of BLKSIZE */
} String;

struct tm_str {        /* used for generating msg-id and date */
  char tm_str_sec[2];  /* "00".."59"                          */
  char tm_str_min[2];  /* "00".."59"                          */
  char tm_str_hour[2]; /* "00".."23"                          */
  char tm_str_day[2];  /* "01".."31"                          */
  char tm_str_mon[2];  /* "01".."12"                          */
  char tm_str_year[4]; /* "1900".."2038"                      */
  char *tm_str_smon;   /* pointer into months[]               */
  char *tm_str_sday;   /* pointer into days[]                 */
};

enum recipient_determination { /* send email to recipients given by...    */
  ARGUMENT_HEADER_IF_NONE,     /* ...arguments, or header if no arguments */
  ARGUMENT_ONLY,               /* ...arguments only                       */
  ARGUMENT_AND_HEADER,         /* ...arguments and header                 */
  HEADER_ONLY,                 /* ...header only                          */
};


char *relay;         /* signal handlers use it                         */
int read_timeout,    /* we have a generic function called w2s which    */
    write_timeout,   /*   we use when talking to the smtp relay        */
    verbosity,       /* 0=quite, 1=info, 2=debug,                      */
    lazy;            /* if zero: die on warnings, strict header checks */


/* "initialize" s by assigning NULLSTRING before calling str_realloc */
void str_realloc (String *s, const size_t new_strlen)
{
  static size_t new_buflen;

  new_buflen = (new_strlen / BLKSIZE) * BLKSIZE + BLKSIZE;
  if (new_buflen > s->_l) {
    s->s = realloc(s->s, new_buflen);
    if (!s->s) {
      write(STDERR, "Fatal error: no memory!\n", 24);
      exit(2);
    }
    s->_l = new_buflen;
  }
}

/* append a zero terminated string of known lenght */
void str_cat_n (String *dst, const char *s, const size_t tmp)
{
  static size_t new_strlen;

  new_strlen = dst->l + tmp;
  str_realloc(dst, new_strlen);
  memmove(dst->s + dst->l, s, tmp);
  *(dst->s + new_strlen) = '\0';
  dst->l = new_strlen;
}

/* append a zero terminated string */
void str_cat (String *dst, const char *s)
{
  str_cat_n(dst, s, strlen(s));
}

/* convert unsigned long of known lenght into string */
void ul2str (char *dst, unsigned long int src, int len)
{
  for (--len; len >= 0; --len) {
    dst[len] = '0' + src % 10;
    src /= 10;
  }
}

/* append an unsigned long */
void str_cat_ul (String *dst, unsigned long int i)
{
  static size_t len;
  unsigned long int save;

  save = i;
  len = 0; 
  do {
    ++len;
    i /= 10;
  } while (i);
  str_realloc(dst, dst->l + len);
  *(dst->s + dst->l + len) = '\0';
  ul2str(dst->s + dst->l, save, len);
  dst->l += len;
}

/* insert a character at position pos (pos=0..s.l-1) */
void str_insert_chr (String *s, size_t pos, char c)
{
  str_realloc(s, s->l + 1);
  memmove(s->s + pos + 1, s->s + pos, s->l - pos);
  *(s->s + pos) = c;
  ++s->l;
}

/* appends now as a well formed string eg. "Wed, 11 Aug 2003 09:41:30 -0000" */
void str_cat_date(String *s, struct tm_str *now)
{
  str_cat_n(s, now->tm_str_sday, 3);
  str_cat_n(s, ", ", 2);
  str_cat_n(s, now->tm_str_day, 2);
  str_cat_n(s, " ", 1);
  str_cat_n(s, now->tm_str_smon, 3);
  str_cat_n(s, " ", 1);
  str_cat_n(s, now->tm_str_year, 4);
  str_cat_n(s, " ", 1);
  str_cat_n(s, now->tm_str_hour, 2);
  str_cat_n(s, ":", 1);
  str_cat_n(s, now->tm_str_min, 2);
  str_cat_n(s, ":", 1);
  str_cat_n(s, now->tm_str_sec, 2);
  str_cat_n(s, " -0000", 6);
}

/* return a struct tm_str containing current date and time as strings */
struct tm_str *get_now ()
{
  static time_t sec_epoch = 0;
  static struct tm_str ret;
  static struct tm *tmp;

  if (sec_epoch) return &ret;
  sec_epoch = time(NULL);
  tmp = gmtime(&sec_epoch);
  ret.tm_str_smon = months[tmp->tm_mon];
  ret.tm_str_sday = days[tmp->tm_wday];
  ++tmp->tm_mon;
  tmp->tm_year += 1900;
  ul2str(ret.tm_str_sec, tmp->tm_sec, 2);
  ul2str(ret.tm_str_min, tmp->tm_min, 2);
  ul2str(ret.tm_str_hour, tmp->tm_hour, 2);
  ul2str(ret.tm_str_day, tmp->tm_mday, 2);
  ul2str(ret.tm_str_mon, tmp->tm_mon, 2);
  ul2str(ret.tm_str_year, tmp->tm_year, 4);
  return &ret;
}

int is_white_space (char c)
{
  return ((c == ' ') || (c == '\r') || (c == '\t') || (c == '\n'));
}

/* print s1 to descriptor d; if not NULL print s2 prepended by ": " too */
void outp (const int d, const char *s1, const char *s2, const char *type)
{
  static String s = NULLSTRING;

  if (type) str_cat(&s, type); /* eg. "Error: ", "Warning: ", blabla */

  str_cat(&s, s1);

  if (s2) {
    str_cat_n(&s, ": ", 2);
    str_cat(&s, s2);
  }
  str_cat_n(&s, "\n", 1);
  write(d, s.s, s.l);
  s.l = 0;
}

/* inform user on stderr about non fatal error; quit program if not lazy */
void warn (const char *s1, const char *s2)
{
  if (!lazy) outp(STDERR, s1, s2, "WARNING: ");
}

/* print out fatal error and quit program */
void die (const char *s1, const char *s2)
{
  outp(STDERR, s1, s2, "ERROR: ");
  exit(1);
}

/* generic help */
void usage ()
{
  const char help[] =
"smtppipe V"VERSION" - reads mail from stdin and forwards it to a smtp relay\n"
"conformal with sendmail and qmail-inject command line switches\n\n"
"smtppipe [-a] [-h] [-t] [-A] [-H] [-f sender] [-m smtprelay] [-p port] [-l] [-x]\n"
"         [-ct connect_timeout] [-rt read_timeout] [-wt write_timeout] [-v]\n"
"         [-oem] [-oi] [--] [recipient ... ]\n\n"
"smtppipe --help\n"
"smtppipe -help\n"
"smtppipe -V\n\n"
"  -a         send message to all addresses given as recipients arguments;\n"
"             do not use any header recipient addresses\n"
"  -h, -t     send message to all header recipient addresses (To, Cc, Bcc);\n"
"             do not use any command line recipients\n"
"  -A         send message to all addresses given as recipient arguments; if\n"
"             no recipient arguments are given, use header recipient addresses;\n"
"             this is the default behaviour\n"
"  -H         send message to all addresses given as recipient arguments and\n"
"             to all addresses given as header recipients\n"
"  -f         forces envelope sender address; if not specified then the\n"
"             From:-line found in the header, the environment parameters USER,\n"
"             QMAILUSER, LOGNAME and the name of the account under which \n"
"             smtppipe runs (in that order) are checked for an user name\n"
"  -m         host name or address of a smtp relay; if not specified then the\n"
"             environment variables SMTPRELAY, QMAILHOST and MAILHOST (in that\n"
"             order) are checked for a host name or an address\n"
"  -p port    use specified port number when connecting to the smtp relay\n"
"             instead of the default SMTP port; if the host name of the smtp\n"
"             relay is given in a host:port notation, then that port is used\n"
"  -l         be lazy; don't print out warnings\n"
"  -x         don't insert missing From, Subject, Message-ID and Date lines\n"
"             into the header of your mail\n"
"  -ct #      timeout on connecting to smtp relay; defaults to 60 seconds\n"
"  -rt #      timeout on reading from smtp relay; defaults to 120 seconds\n"
"  -wt #      timeout on writing to smtp relay; defaults to 120 seconds\n"
"  -v         be verbose (print out what is being done);\n"
"             use -vv for debugging;\n"
"  -oem       not handled due to compatibility\n"
"  -oi        not handled due to compatibility\n"
"  --         not handled due to compatibility\n"
"  recipient  list of space separated recipient addresses\n"
"  --help     displays this help\n"
"  -help      displays this help\n"
"  -V         displays version\n\n"
"smtppipe supports authentication against the smtp relay only through the\n"
"following environment parameters:\n\n"
"  $SMTPUSER  cleartext user name used for authentication\n"
"  $SMTPPASS  cleartext password used for authentication; both $SMTPUSER\n"
"             and $SMTPPASS must be set for authentication to happen.\n"
"  $SMTPMECH  force authentication method; if not set then smtppipe will try to\n"
"             use LOGIN or PLAIN (in that order) if supported by the smtp relay\n\n"
"smtppipe handles bcc recipients and deletes them before final delivery.\n\n"
"see http://www.ogris.de/smtppipe/\n";

  write(STDOUT, help, strlen(help));
  exit(0);
}

/* signal handlers used when talking to the smtp relay */
void read_timeout_handler () { die(relay, "read timeout"); }
void write_timeout_handler () { die(relay, "write timeout"); }
void connect_timeout_handler () { die(relay, "connect timeout"); }

/* set the specified signal handler */
void set_alarm (int timeout, void *handler)
{
  static struct sigaction sa;

  memset(&sa, 0, sizeof(struct sigaction));
  if (handler) sa.sa_handler = handler;
  else sa.sa_handler = SIG_DFL;
  if (sigaction(SIGALRM, &sa, NULL)) die("sigaction()", strerror(errno));
  alarm(timeout);
}

/* write command to socket; if specified write value prepended by ": " too    */
/* if specified check whether smtp relay returns the desired code; die if not */
char *w2s (int socket, char *command, char *value, char *code)
{
  static String s = NULLSTRING, buffer = NULLSTRING, err = NULLSTRING;
  static ssize_t ret;
  static size_t l;
  static int cont;

  s.l = 0;
  if (command) {
    str_cat(&s, command);
    if (value) str_cat(&s, value);
    str_cat_n(&s, "\r\n", 2);

    DEBUGP("To smtp relay", s.s);
    set_alarm(write_timeout, write_timeout_handler);
    ret = write(socket, s.s, s.l);
    set_alarm(0, NULL);

    if (ret < 0) die("write(socket)", strerror(errno));
    if ((size_t) ret != s.l)
      die("write(socket)", "could not write whole data stream");
  }

  if (!code) return "";

  buffer.l = 0;
  cont = 1;
  do {
    set_alarm(read_timeout, read_timeout_handler);
    str_realloc(&buffer, buffer.l + BLKSIZE);
    ret = read(socket, buffer.s + buffer.l, BLKSIZE);
    if (ret < 0) die("read(socket)", strerror(errno));
    buffer.l += ret;
    l = buffer.l;
    while (l > 3) {
      --l;
      if (*(buffer.s + buffer.l - 1) != '\n') break;
      if (l > 4) if (*(buffer.s + l - 4) != '\n') continue;
      if (*(buffer.s + l) == ' ') { cont = 0; break; }
      if (*(buffer.s + l) == '-') break;
    }
  } while (cont);
  set_alarm(0, NULL);

  *(buffer.s + buffer.l) = '\0';
  DEBUGP("From smtp relay", buffer.s);

  if (!strncmp(buffer.s, code, strlen(code))) return buffer.s;

  /* smtp relay did not return our desired status code => quit program */
  if (buffer.l) {
    while ((*(buffer.s + buffer.l - 1) == '\r') ||
           (*(buffer.s + buffer.l - 1) == '\n')) --buffer.l;
    str_cat(&err, "The smtp relay returns an unexpected status code:\n");
    str_cat_n(&err, buffer.s, buffer.l);
  }
  else str_cat(&err, "The smtp relay returns nothing.");
  if (command && strcmp(code, "334") && strcmp(code, "235")) {
    str_cat(&err, "\nThe requested command was: ");
    str_cat_n(&err, s.s, s.l - 2);
  }
  str_cat(&err, "\nThe expected status code was");
  die(err.s, code);
  /* not reached, make compiler happy */
  return NULL;
}

/* try to get a host name or an ip address of the smtp relay; else die */
/* this can be easily extended to fit your environment                 */
char *get_relay ()
{
  char *relay;

  relay = getenv("SMTPRELAY");
  DEBUGP("SMTPRELAY", relay);
  if (!relay) relay = getenv("QMAILHOST");
  DEBUGP("QMAILHOST", relay);
  if (!relay) relay = getenv("MAILHOST");
  DEBUGP("MAILHOST", relay);
  if (!relay) die("no smtprelay specified", NULL);
  return relay;
}

/* try to get a user name which can be used as sender; else die */
/* this can be easily extended to fit your environment          */
char *get_user (const char *hostname)
{
  String s = NULLSTRING;
  struct passwd *pw;
  char *user;

  user = getenv("USER");
  DEBUGP("USER", user);
  if (!user) user = getenv("QMAILUSER");
  DEBUGP("QMAILUSER", user);
  if (!user) user = getenv("LOGNAME");
  DEBUGP("LOGNAME", user);
  if (!user) {
    pw = getpwuid(getuid());
    if (pw) user = pw->pw_name;
  }
  DEBUGP("getpwuid()", user);
  if (!user) die("can not figure out your sender's address", NULL);
  str_cat(&s, user);
  str_cat_n(&s, "@", 1);
  str_cat(&s, hostname);
  return s.s;
}

/* try to get a user name which can be used for authentication */
char *get_smtpuser ()
{
  return getenv("SMTPUSER");
}

/* try to get a password for authentication */
char *get_smtppass ()
{
  return getenv("SMTPPASS");
}

/* try to get a authentication method */
char *get_smtpmech ()
{
  return getenv("SMTPMECH");
}

char *b64enc (char *c)
{
  static const char tbl[65] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                              "abcdefghijklmnopqrstuvwxyz0123456789+/";
  static String ret = NULLSTRING;
  static size_t len, i, j;

  len = strlen(c);
  str_realloc(&ret, 4 * (len + 2) / 3);
  for (i = 0, j = 0; i < len; ++i) {
    *(ret.s + j++) = tbl[(*(c + i) >> 2) & 63];
    *(ret.s + j++) = tbl[((*(c + i) << 4) + (*(c + i + 1) >> 4)) & 63];
    if (++i == len) break;
    *(ret.s + j++) = tbl[((*(c + i) << 2) + (*(c + i + 1) >> 6)) & 63];
    if (++i == len) break;
    *(ret.s + j++) = tbl[*(c + i) & 63];
  }
  while ((j % 4)) *(ret.s + j++) = '=';
  *(ret.s + j) = '\0';
  return ret.s;
}

void auth_login (int socket, char *smtpuser, char *smtppass, char *hostname)
{
  hostname = hostname;
  w2s(socket, "AUTH LOGIN", NULL, "334");
  w2s(socket, b64enc(smtpuser), NULL, "334");
  w2s(socket, b64enc(smtppass), NULL, "235");
}

void auth_plain (int socket, char *smtpuser, char *smtppass, char *hostname)
{
  static String auth = NULLSTRING;

  hostname = hostname;
  auth.l = 0;
  str_cat(&auth, smtpuser);
  str_cat_n(&auth, "\0", 1);
  str_cat(&auth, smtpuser);
  str_cat_n(&auth, "\0", 1);
  str_cat(&auth, smtppass);
  w2s(socket, "AUTH PLAIN", NULL, "334");
  w2s(socket, b64enc(auth.s), NULL, "235");
}

void authenticate (int socket, char *smtpuser, char *smtppass, char *smtpmech,
                   char *hostname)
{
  struct mechs_list {
    char *name;
    void (*worker)(int, char*, char*, char*);
  } mechs[] = {
    { "LOGIN", auth_login },
    { "PLAIN", auth_plain },
  };
  char *p;
  int i = 0, cnt = sizeof(mechs) / sizeof(struct mechs_list);

  if (smtpmech || !smtpuser || !smtppass) {
    /* simple greeting */
    w2s(socket, "HELO ", hostname, "250");
    if (!smtpuser || !smtppass) {
      DEBUGP("no smtpuser nor smtppass", NULL);
      return;
    }
    for (i = 0; i < cnt; ++i) if (!strcmp(smtpmech, mechs[i].name)) break;
    if (i == cnt) die("authentication method not supported", smtpmech);
  }
  else {
    char *ehlo = w2s(socket, "EHLO ", hostname, "250");
    smtpmech = strstr(ehlo, "\n250-AUTH ");
    if (!smtpmech) smtpmech = strstr(ehlo, "\n250 AUTH ");
    if (!smtpmech)
      die("The smtp relay does not offer any authentication methods", ehlo);
    for (smtpmech += 9; *smtpmech == ' '; smtpmech = p) {
      for (p = ++smtpmech; !is_white_space(*p); ++p) if (!*p) break;
      for (i = 0; i < cnt; ++i)
        if (!strncmp(smtpmech, mechs[i].name, p - smtpmech)) break;
      if (i < cnt) break;
    }
    if (i == cnt)
      die("The smtp relay does not offer any supported authentication methods",
          ehlo);
  }
  INFOP("using auth mech", mechs[i].name);
  mechs[i].worker(socket, smtpuser, smtppass, hostname);
}

/* strip leading '<', append trailing '>' */
char *only_leading_angle (char *address)
{
  static String addr = NULLSTRING;

  str_cat(&addr, (*address == '<')?(address + 1):(address));
  if (*(addr.s + addr.l - 1) != '>') str_cat_n(&addr, ">", 1);
  addr.l = 0;

  return addr.s;
}

/* construct message-id from date, time and process id and write it to socket */
void msgid2socket (int socket, char *hostname, pid_t pid)
{
  /* Message-Id: <20030811090830.14846.smtppipe@HOSTNAME> */
  static String msgid = NULLSTRING;
  static struct tm_str *now;

  now = get_now();
  str_cat_n(&msgid, "Message-ID: <", 13);
  str_cat_n(&msgid, now->tm_str_year, 4);
  str_cat_n(&msgid, now->tm_str_mon, 2);
  str_cat_n(&msgid, now->tm_str_day, 2);
  str_cat_n(&msgid, now->tm_str_hour, 2);
  str_cat_n(&msgid, now->tm_str_min, 2);
  str_cat_n(&msgid, now->tm_str_sec, 2);
  str_cat_n(&msgid, ".", 1);
  str_cat_ul(&msgid, pid);
  str_cat_n(&msgid, ".smtppipe@", 10);
  str_cat(&msgid, hostname);
  str_cat_n(&msgid, ">", 1);
  INFOP("Generated Message-ID", msgid.s);
  w2s(socket, msgid.s, NULL, NULL);
}

/* write current time and date to socket */
void date2socket (int socket)
{
  static String date = NULLSTRING;

  str_cat_n(&date, "Date: ", 6);
  str_cat_date(&date, get_now());
  INFOP("Generated Date", date.s);
  w2s(socket, date.s, NULL, NULL);
}

/* write a Received line to socket */
void received2socket (int socket, pid_t pid)
{
  static String recv = NULLSTRING;

  str_cat(&recv, "Received: (smtppipe V" VERSION " ");
  str_cat_ul(&recv, pid);
  str_cat(&recv, " invoked by uid ");
  str_cat_ul(&recv, getuid());
  str_cat_n(&recv, ");\r\n\t", 5);
  str_cat_date(&recv, get_now());
  INFOP("Generated Received", recv.s);
  w2s(socket, recv.s, NULL, NULL);
}

/* smtp envelope: write address of sender to socket */
void from2socket (int socket, char *from)
{
  INFOP("From", from);
  w2s(socket, "MAIL FROM:<", only_leading_angle(from), "250");
}

/* smtp envelope: write address of a recipient to socket */
void rcpt2socket (int socket, char *rcpt)
{
  INFOP("Recipient", rcpt);
  w2s(socket, "RCPT TO:<", only_leading_angle(rcpt), "250");
}

/* walks through s until token is found, returns pointer to it or NULL */
char *goto_next_unescaped (char token, char *s)
{
  do {
    ++s;
  } while (*s && ((*s != token) || (*(s - 1) == '\\')));
  return s;
}

/* die if header_this_seen greater zero */
void handle_generic_line (int *header_this_seen, char *this)
{
  DEBUGP("handle_generic_line", this);
  if (*header_this_seen)
    warn(this, "must not appear twice times within a header!");
  ++*header_this_seen;
}

/* parse an address block (From:, To:, Cc:, Bcc:) in the mail header and */
/*   search for valid mail addresses;                                    */
/* die if mail header contains two or more address blocks of same type   */
/* if not parsing a From: line then return if mode permits using         */
/*   header recipients                                                   */
/* die if there fatal typos (missing brackets or quote signs) or extra   */
/*   whitespace between addresses separated by commas                    */

/* "firstname lastname" <foo@bar>
 * nickname <foo@bar>
 * foo@bar (firstname lastname)
 * foo@bar
 */
void handle_address_line (int socket, char *left,
                          enum recipient_determination mode,
                          int first_cmd_line_recip,
                          char **from, char *hostname,
                          int *header_this_seen, char *this)
{
  static char *addrleft, *addrright, save;

  handle_generic_line(header_this_seen, this);

  /* check if we have to handle this */
  if ( ((*this != 'F') && ((mode == ARGUMENT_ONLY) ||
        ((mode == ARGUMENT_HEADER_IF_NONE) && (first_cmd_line_recip > 0)))) ||
       ((*this == 'F') && (*from)) )  {
    DEBUGP("handle_address_line called but not needed; line type", this);
    return;
  }

  for (addrleft = left; *addrleft; addrleft = addrright) {

    /* strip whitespace */
    while (is_white_space(*addrleft)) ++addrleft;
    if (!addrleft) die("line contains only whitespace", left);

    if (*addrleft == '"') {
      DEBUGP(addrleft, " is like \"Firstname Surname\" <user@domain>");
      addrleft = goto_next_unescaped('<', addrleft);
      if (!addrleft) die("line contains no address", left);
      addrright = goto_next_unescaped('>', addrleft);
      if (!addrright) die("unmatched bracket", addrleft);
      ++addrright;
    }
    else {
      addrright = addrleft;
      do {
        ++addrright;
      } while ((*addrright != ' ') && (*addrright != '\r') && (*addrright) &&
               (strncmp(addrright, ", ", 2)) &&   /* commas are allowed */
               (strncmp(addrright, ",\r", 2)));   /*  in mail addresses */

      if (addrright) {
        if (*(addrright + 1) == '<') {
          DEBUGP(addrleft, " is like Nickname <user@domain>");
          addrleft = ++addrright;
          addrright = goto_next_unescaped('>', addrleft);
          if (!addrright) die("unmatched bracket", addrleft);
          ++addrright;
        }
      }
    }

    save = *addrright;
    *addrright = '\0';
    if (*this == 'F') {
      if (*from) die("your mail contains at least two sender addresses", left);
      *from = addrleft;
      from2socket(socket, *from);
    }
    else {
      /* smtp requires a sender before any recipient */
      if (!*from) {
        *from = get_user(hostname);
        from2socket(socket, *from);
      }
      DEBUGP(this, addrleft);
      rcpt2socket(socket, addrleft);
    }
    *addrright = save;
    if (save == ' ') {
      DEBUGP(addrleft, " is like user@domain (Firstname Surname)");
      addrright = goto_next_unescaped(')', addrright);
      if (!addrright) die("unmachted bracket", addrleft);
      ++addrright;
    }
    switch (*addrright) {
      case '\0': /* right is reached, stop processing             */
        return;
      case '\r': /* continuation on next line guarented by main() */
      case ',':  /* next address follows on this line             */
        ++addrright;
        break;
      default:   /* illegal token                                 */
        die("illegal character found in address line", addrleft);
    } /* switch */
  } /* for */
}

/* if cmd line argument at position pos matches test, then set set_on_found */
/*   to the value of that cmd line argument                                 */
int check_arg (char **argv, int *pos, char *test, char **set_on_found)
{
  static int pos_save;
  static size_t len;

  len = strlen(test);
  if (!strncmp(argv[*pos], test, len)) {
    pos_save = *pos;
    if (strlen(argv[*pos]) > len) *set_on_found = argv[*pos] + len;
    else *set_on_found = argv[++*pos];
    if (!*set_on_found || (**set_on_found == '-'))
      die("invalid parameter", argv[pos_save]);
    return 1;
  }
  return 0;
}

/* main routine, access environment via getenv() */
int main (int argc, char *argv[])
{
  struct hostent *he;
  struct sockaddr_in sin;
  size_t len;
  ssize_t ret;
  pid_t pid;
  enum recipient_determination mode;
  int i, sock, first_cmd_line_recip, connect_timeout,
      header_from_seen, header_to_seen, header_cc_seen, header_bcc_seen,
      header_date_seen, header_msgid_seen, header_subject_seen,
      dont_insert_missing_headers;
  char *p, *sep, *left, *right, *from;
  char buf[BLKSIZE];
  String mail = NULLSTRING;
  unsigned short port;

  /* initialization */
  first_cmd_line_recip = 0;
  header_from_seen = 0;
  header_to_seen = 0;
  header_cc_seen = 0;
  header_bcc_seen = 0;
  header_date_seen = 0;
  header_msgid_seen = 0;
  header_subject_seen = 0;
  dont_insert_missing_headers = 0;
  mode = ARGUMENT_HEADER_IF_NONE;
  from = NULL;
  relay = NULL;
  port = SMTP_PORT;
  connect_timeout = CONNECT_TIMEOUT;
  read_timeout = READ_TIMEOUT;
  write_timeout = WRITE_TIMEOUT;
  verbosity = 0;
  lazy = 0;

  /* parse command line arguments */
  for (i = 1; i < argc; ++i) {
    if ( (!strcmp(argv[i], "--help")) || (!strcmp(argv[i], "-help")) ) usage();

    if (!strcmp(argv[i], "-V")) {
      write(STDOUT, "smtppipe V" VERSION "\n", 15);
      return 0;
    }

    if (check_arg(argv, &i, "-f", &from)) continue;   /* from                */
    if (check_arg(argv, &i, "-m", &relay)) continue;  /* relay               */
    if (strncmp(argv[i], "-", 1)) {
      /* if argument does not begin with a '-' it's assumed to be an address */
      first_cmd_line_recip = i;
      break;
    }

    if (!strcmp(argv[i], "--")) continue;             /* mutt                */
    if (!strcmp(argv[i], "-oem")) continue;           /* emacs uses -oem -oi */
    if (!strcmp(argv[i], "-oi")) continue;            /* emacs uses -oem -oi */

    if (check_arg(argv, &i, "-p", &p)) {              /* port                */
      port = atoi(p);
      continue;
    }
    if (check_arg(argv, &i, "-ct", &p)) {             /* connect timeout     */
      connect_timeout = atoi(p);
      continue;
    }
    if (check_arg(argv, &i, "-rt", &p)) {             /* read timeout        */
      read_timeout = atoi(p);
      continue;
    }
    if (check_arg(argv, &i, "-wt", &p)) {             /* write timeout       */
      write_timeout = atoi(p);
      continue;
    }

    if (!strcmp(argv[i], "-a")) {
      mode = ARGUMENT_ONLY;
      continue;
    }

    if ( (!strcmp(argv[i], "-h")) || (!strcmp(argv[i], "-t")) ) {
      mode = HEADER_ONLY;
      continue;
    }

    if (!strcmp(argv[i], "-A")) {
      mode = ARGUMENT_HEADER_IF_NONE;
      continue;
    }

    if (!strcmp(argv[i], "-H")) {
      mode = ARGUMENT_AND_HEADER;
      continue;
    }

    if (!strncmp(argv[i], "-v", 2)) {
      verbosity = strlen(argv[i]) - 1;
      continue;
    }

    if (!strcmp(argv[i], "-l")) {
      lazy = 1;
      continue;
    }

    if (!strcmp(argv[i], "-x")) {
      dont_insert_missing_headers = 1;
      continue;
    }

    warn("unknown parameter", argv[i]);
  } /* parsing cmd line arguments */

  /* must have an smtprelay */
  if (!relay) relay = get_relay();
  DEBUGP("Relay", relay);
  p = strchr(relay, ':');
  if (p) {
    *p++ = '\0';
    port = atoi(p);
  }
  if (verbosity >= 2) {
    String foo = NULLSTRING;
    str_cat_ul(&foo, port);
    DEBUGP("Port", foo.s);
  }
  if (!(he = gethostbyname(relay))) die(relay, hstrerror(h_errno));
  INFOP("FQDN of smtp relay", he->h_name);

  memcpy((char*) &sin.sin_addr, he->h_addr, he->h_length);
  sin.sin_port = htons(port);
  sin.sin_family = PF_INET;

  /* determine local hostname to do the initial SMTP HELO and to form */
  /* the sender's address if not specified otherwise                  */
  if (gethostname(buf, BLKSIZE)) die("gethostname()", strerror(errno));
  DEBUGP("My hostname", buf);
  if (!(he = gethostbyname(buf))) die(buf, hstrerror(h_errno));
  INFOP("My FQDN", he->h_name);

  /* get the mail through STDIN */
  for (;;) {
    ret = read(STDIN, buf, BLKSIZE);
    if (!ret) break;
    if (ret < 0) die("read(STDIN)", strerror(errno));
    str_cat_n(&mail, buf, ret);
  }

  /* empty mail */
  if (!mail.l) die("nothing to do", "empty mail");

  /* count newline w/o prepended '\r' */
  for (len = 0, p = mail.s + 1; *p; ++p) {
    if (*p != '\n') continue;
    if (*(p - 1) == '\r') continue;
    ++len;
  }
  str_realloc(&mail, mail.l + len);
  p = mail.s + mail.l - 1;
  mail.l += len;
  /* replace each single '\n' by "\r\n" */
  for (i = 1; (p > mail.s) && len; --p, ++i) {
    if (*p != '\n') continue;
    if (*(p - 1) == '\r') continue;
    memmove(p + len, p, i);
    i = 0;
    --len;
    *(p + len) = '\r';
  }

  /* no incomplete last line */
  if (*(mail.s + mail.l - 1) ==  '\r') str_cat_n(&mail, "\n", 1);
  else if (strcmp(mail.s + mail.l - 2, "\r\n")) str_cat_n(&mail, "\r\n", 2);

  /* escape single dots */
  for (len = 0; len < mail.l; ++len) {
    p = mail.s + len;
    if (strncmp(p, "\n.\r", 3)) continue;
    len += 3;
    str_insert_chr(&mail, len, '.');
  }

  /* try to find a body separation */
  sep = strstr(mail.s, "\r\n\r\n");
  if (!sep) DEBUGP("No body sepration found", NULL);

  /* connect to server */
  if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    die("socket()", strerror(errno));

  set_alarm(connect_timeout, connect_timeout_handler);
  if (connect(sock, (struct sockaddr*) &sin, sizeof(sin)))
    die("connect()", strerror(errno));
  set_alarm(0, NULL);

  /* smtp blabla */
  w2s(sock, NULL, NULL, "220");
  authenticate(sock, get_smtpuser(), get_smtppass(), get_smtpmech(),
               he->h_name);
  if (from) from2socket(sock, from);

  /* look for header from address, evaluate header recips, erase bcc's */
  /* but only if we have an empty line that separates header and body  */
  if (sep) for (left = mail.s; left < sep;) {
    /* left points to the beginning of a header line                 */
    /* right points to the end of a (possibly continued) header line */

    for (right = left;; right += 2) {
      right = strchr(right, '\r');       /* cant be NULL because sep != NULL */
      if ((*(right + 2) != ' ') &&
          (*(right + 2) != '\t')) break; /* line is not continued            */
    }
    *right = '\0';                       /* for parsing this line            */

    if (!strncasecmp(left, "bcc:", 4)) {
      /* bcc needs special treatment */
      DEBUGP("Bcc:-line", left);
      handle_address_line(sock, left + 4, mode,
                          first_cmd_line_recip, &from, he->h_name,
                          &header_bcc_seen, "Bcc");
      /* erase bcc line */
      right += 2;
      memmove(left, right, mail.l - (right - mail.s));
      mail.l -= right - left;
      sep -= right - left;
    }

    else {

      /* from */
      if (!strncasecmp(left, "from:", 5))
        handle_address_line(sock, left + 5, mode,
                            first_cmd_line_recip, &from, he->h_name,
                            &header_from_seen, "From");

      /* to */
      else if (!strncasecmp(left, "to:", 3))
        handle_address_line(sock, left + 3, mode,
                            first_cmd_line_recip, &from, he->h_name,
                            &header_to_seen, "To");

      /* cc */
      else if (!strncasecmp(left, "cc:", 3))
        handle_address_line(sock, left + 3, mode,
                            first_cmd_line_recip, &from, he->h_name,
                            &header_cc_seen, "Cc");

      /* date */
      else if (!strncasecmp(left, "date:", 5))
        handle_generic_line(&header_date_seen, "Date");

      /* message-id */
      else if (!strncasecmp(left, "message-id:", 11))
        handle_generic_line(&header_msgid_seen, "Message-ID");

      /* subject */
      else if (!strncasecmp(left, "subject:", 8))
        handle_generic_line(&header_subject_seen, "Subject");

      *right = '\r';    /* remove zero termination */
      left = right + 2; /* jump to next line       */
    }
  } /* for */

  /* evaluate cmd line recipients */
  if ((first_cmd_line_recip > 0) &&
      ((mode == ARGUMENT_HEADER_IF_NONE) ||
       (mode == ARGUMENT_ONLY) ||
       (mode == ARGUMENT_AND_HEADER))) {
    DEBUGP("argument recipients...", NULL);
    if (!from) {
      from = get_user(he->h_name);
      from2socket(sock, from);
    }
    for (i = first_cmd_line_recip; i < argc; i++) {
      DEBUGP("argument recipient", argv[i]);
      rcpt2socket(sock, argv[i]);
    }
  }

  /* no recipients at all? => quit program */
  if (!first_cmd_line_recip && !header_to_seen &&
      !header_cc_seen && !header_bcc_seen) 
    die("no recipients found at all", NULL);

  /* send the beast */
  str_cat_n(&mail, ".", 1);
  w2s(sock, "DATA", NULL, "354");

  /* insert a Received: line */
  pid = getpid();
  received2socket(sock, pid);

  /* insert unseen header lines */
  if (!dont_insert_missing_headers) {
    if (!header_from_seen) {
      if (*from == '<') ++from;
      if (*(from + strlen(from) - 1) == '>') *(from + strlen(from)) = '\0';
      INFOP("Generated From", from);
      w2s(sock, "From: ", from, NULL);
    }
    if (!header_subject_seen) {
      INFOP("Generated empty Subject", NULL);
      w2s(sock, "Subject: ", "<no subject given>", NULL);
    }
    if (!header_msgid_seen) msgid2socket(sock, he->h_name, pid);
    if (!header_date_seen) date2socket(sock);
    if (!sep) {
      INFOP("Generated body separation", NULL);
      w2s(sock, "", NULL, NULL);
    }
  }

  DEBUGP("sending mail...", NULL);
  w2s(sock, mail.s, NULL, "250");
  w2s(sock, "QUIT", NULL, "221");
  close(sock);

  DEBUGP("exiting...", NULL);
  return 0;
}
